use std::{time::{SystemTime, UNIX_EPOCH}, io::Write};

use reqwest::Client;
use serde::Deserialize;
use serde::{Serialize, de::DeserializeOwned};
use serde_json::json;
use clap::ArgEnum;

use crate::crypto::{gen_signed_req, get_session_key, get_track_key, decrypt_frame};
use crate::mp4::{parse_segment, BoxType, ExtBoxType};

const API_URL: &str = "https://www.qobuz.com/api.json/0.2/";

const APP_ID: u32 = 950096963;

const USER_AGENT: &str = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.51 Safari/537.36";

#[derive(Clone, ArgEnum, Debug)]
pub enum Method {
    GET,
    POST
}

pub struct APIClient {
    client: Client,
    auth_token: Option<String>,
    session: Session
}

impl APIClient {
    pub fn new() -> Self {
        let client = Client::builder()
            .user_agent(USER_AGENT)
            .build()
            .unwrap();

        Self {
            client,
            auth_token: None,
            session: Session {
                id: String::new(),
                key: [0u8; 16],
                expiry: 0
            }
        }
    }

    pub fn set_auth_token(&mut self, auth_token: &str) {
        self.auth_token = Some(auth_token.to_string());
    }

    pub async fn api_call<T, R>(&self, object: &str, method: &str, params: &T, req_method: Method, signed: bool) -> Result<R, reqwest::Error>
    where T: Serialize + ?Sized,
          R: DeserializeOwned {
        let url = format!("{API_URL}{object}/{method}");
        let reqwest_method = match req_method {
            Method::GET => reqwest::Method::GET,
            Method::POST => reqwest::Method::POST
        };

        let mut req = self.client.request(reqwest_method, url)
            .header("x-app-id", APP_ID);

        if let Some(token) = &self.auth_token {
            req = req.header("x-user-auth-token", token);
        }

        match req_method {
            Method::GET => {
                if signed {
                    req = req.query(&gen_signed_req(object, method, params));
                }
                req = req.query(params);
            },
            Method::POST => {
                let mut form_data = String::new();
                if signed {
                    form_data.push_str(&serde_urlencoded::to_string(&gen_signed_req(object, method, params)).unwrap());
                }
                let params_form = serde_urlencoded::to_string(params).unwrap();
                if !params_form.is_empty() {
                    form_data.push('&');
                    form_data.push_str(&params_form);
                }
                req = req.body(form_data)
                    .header("content-type", "application/x-www-form-urlencoded");
            }
        }

        if object == "file" && method == "url" {
            req = req.header("x-session-id", &self.session.id);
        }

        let resp = req.send().await?
            .error_for_status()?;

        resp.json().await
    }

    pub async fn renew_session(&mut self) {
        let timestamp = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .unwrap()
            .as_secs();
        
        if timestamp >= self.session.expiry {
            let resp: SessionResp = self.api_call("session", "start", &json!({"profile":"qbz-1"}), Method::POST, true).await.unwrap();

            self.session.id = resp.session_id;
            get_session_key(&resp.infos, &mut self.session.key);
            self.session.expiry = resp.expires_at;
        }
    }

    pub async fn get_track_url(&mut self, track_id: u32, format_id: u32) -> Result<TrackUrl, reqwest::Error> {
        self.renew_session().await;
        
        let params = json!(
            {
                "track_id": track_id,
                "format_id": format_id,
                "intent": "stream"
            }
        );

        let resp: TrackUrlResp = self.api_call("file", "url", &params, Method::GET, true).await?;
        let mut key = [0u8; 16];
        get_track_key(&resp.key, &self.session.key, &mut key);

        Ok(TrackUrl {
            url_template: resp.url_template,
            key,
            n_segments: resp.n_segments
        })
    }

    pub async fn dl_track<T>(&self, url_template: &str, key: &[u8; 16], n_segments: u64, out: &mut T) -> Result<(), reqwest::Error>
    where T: Write {
        for seg in 0..n_segments + 1 {
            let resp = self.client.get(url_template.replace("$SEGMENT$", &seg.to_string()))
                .send().await?
                .error_for_status()?;

            let block = resp.bytes().await?;

            let boxes = parse_segment(&block);
            println!("segment {}", seg);
            //println!("{:#?}", boxes);
            for atom in &boxes {
                println!("box size {}", atom.box_data.len());
                if let BoxType::ExtendedBox {ext_box_type, ..} = &atom.box_type {
                    match ext_box_type {
                        ExtBoxType::Init { initial_data, .. } => {
                            out.write_all(initial_data).unwrap();
                        },
                        ExtBoxType::Segment {frames, ..} => {
                            for frame in frames {
                                if frame.flags != 0 {
                                    let frame_data = decrypt_frame(&frame.frame_data, &key, &frame.iv_raw);
                                    out.write_all(&frame_data).unwrap();
                                } else {
                                    out.write_all(frame.frame_data).unwrap();
                                }
                            }
                        }
                    }
                }
            }
            /*for b in boxes {
                out.write_all(b.box_data).unwrap();
            }*/
        };

        Ok(())
    }
}

#[derive(Deserialize)]
struct TrackUrlResp {
    url_template: String,
    key: String,
    n_segments: u64
}

pub struct TrackUrl {
    pub url_template: String,
    pub key: [u8; 16],
    pub n_segments: u64
}

#[derive(Deserialize)]
struct SessionResp {
    session_id: String,
    infos: String,
    expires_at: u64
}

struct Session {
    id: String,
    key: [u8; 16],
    expiry: u64
}