use std::io::{Cursor, Read};
use byteorder::{BigEndian, ReadBytesExt};

#[derive(Debug)]
pub struct Box<'a> {
    pub box_data: &'a [u8],
    pub box_type: BoxType<'a>,
}

#[derive(Debug)]
pub enum BoxType<'a> {
    Box { box_type: String },
    ExtendedBox { ext_box_type: ExtBoxType<'a>, version: u8, flags: u32 },
}

#[derive(Debug)]
pub enum ExtBoxType<'a> {
    Init {
        track_id: u32,
        file_id: u32,
        sample_rate: u32,
        bits_per_sample: u8,
        channels_count: u8,
        samples_count: u64,
        initial_data_size: u16,
        initial_data: &'a [u8],
        key_id_size: u8,
        key_id: &'a [u8],
        segments_count: u16,
        segment_infos: Vec<SegmentInfo>
    },
    Segment {
        base_offset: u32,
        iv_size: u8,
        n_frames: u32,
        frames: Vec<Frame<'a>>
    }
}

#[derive(Debug)]
pub struct SegmentInfo {
    pub length: u32,
    pub samples_count: u32
}

#[derive(Debug)]
pub struct Frame<'a> {
    pub length: u32,
    pub samples_count: u16,
    pub flags: u16,
    pub iv_raw: &'a [u8],
    pub frame_data: &'a [u8]
}

pub fn parse_segment(buf: &[u8]) -> Vec<Box> {
    let mut reader = Cursor::new(buf);
    let mut boxes: Vec<Box> = Vec::new();

    while reader.position() < buf.len() as u64 {
        let size = reader.read_u32::<BigEndian>().unwrap();

        let box_start = reader.position() - 4;
        let box_end = box_start + size as u64;

        let box_data = &buf[reader.position() as usize..box_end as usize];
        let mut box_reader = Cursor::new(box_data);

        let mut box_type = [0u8; 4];
        box_reader.read_exact(&mut box_type).unwrap();
        let box_type = String::from_utf8(box_type.to_vec()).unwrap();

        let box_type_enum;

        if box_type == "uuid" {
            let uuid = box_reader.read_u128::<BigEndian>().unwrap();
            let uuid = format!("{:016x}", uuid);
            let version = box_reader.read_u8().unwrap();
            let flags = box_reader.read_u24::<BigEndian>().unwrap();

            let ext_box_data = &box_data[box_reader.position() as usize..];

            let ext_box_type = match uuid.as_str() {
                "c7c75df0fdd951e98fc22971e4acf8d2" => parse_init_box(ext_box_data),
                "3b42129256f35f75923663b69a1f52b2" => parse_segment_box(ext_box_data, box_start, buf),
                _ => panic!("Unknown extended box type: {}", uuid)
            };

            box_type_enum = BoxType::ExtendedBox {
                ext_box_type,
                version,
                flags
            };
        } else {
            box_type_enum = BoxType::Box {
                box_type
            };
        }

        boxes.push(Box {
            box_data,
            box_type: box_type_enum
        });

        reader.set_position(box_end);
    }

    boxes
}

fn parse_init_box<'a>(buf: &'a [u8]) -> ExtBoxType<'a> {
    let mut reader = Cursor::new(buf);

    let track_id = reader.read_u32::<BigEndian>().unwrap();
    let file_id = reader.read_u32::<BigEndian>().unwrap();
    let sample_rate = reader.read_u32::<BigEndian>().unwrap();
    let bits_per_sample = reader.read_u8().unwrap();
    let channels_count = reader.read_u8().unwrap();
    let samples_count = reader.read_u64::<BigEndian>().unwrap();
    let initial_data_size = reader.read_u16::<BigEndian>().unwrap();
    let initial_data = &buf[reader.position() as usize..reader.position() as usize + initial_data_size as usize];
    reader.set_position(reader.position() + initial_data_size as u64);
    let key_id_size = reader.read_u8().unwrap();
    let key_id = &buf[reader.position() as usize..reader.position() as usize + key_id_size as usize];
    reader.set_position(reader.position() + key_id_size as u64);
    let segments_count = reader.read_u16::<BigEndian>().unwrap();

    let mut segment_infos: Vec<SegmentInfo> = Vec::new();
    for _ in 0..segments_count {
        segment_infos.push(SegmentInfo {
            length: reader.read_u32::<BigEndian>().unwrap(),
            samples_count: reader.read_u32::<BigEndian>().unwrap()
        });
    }

    ExtBoxType::Init {
        track_id,
        file_id,
        sample_rate,
        bits_per_sample,
        channels_count,
        samples_count,
        initial_data_size,
        initial_data,
        key_id_size,
        key_id,
        segments_count,
        segment_infos
    }
}

fn parse_segment_box<'a>(buf: &'a [u8], box_start: u64, seg_buf: &'a [u8]) -> ExtBoxType<'a> {
    let mut reader = Cursor::new(buf);

    let base_offset = reader.read_u32::<BigEndian>().unwrap();
    let iv_size = reader.read_u8().unwrap();
    let n_frames = reader.read_u24::<BigEndian>().unwrap();

    let mut frames: Vec<Frame<'a>> = Vec::new();
    let mut frame_start = box_start as usize + base_offset as usize;
    for _ in 0..n_frames {
        let length = reader.read_u32::<BigEndian>().unwrap();
        let samples_count = reader.read_u16::<BigEndian>().unwrap();
        let flags = reader.read_u16::<BigEndian>().unwrap();
        let iv_raw = &buf[reader.position() as usize..reader.position() as usize + iv_size as usize];
        reader.set_position(reader.position() + iv_size as u64);

        let frame_end = frame_start + length as usize;

        let frame_data = &seg_buf[frame_start..frame_end];

        frames.push(Frame {
            length,
            samples_count,
            flags,
            iv_raw,
            frame_data
        });

        frame_start = frame_end;
    }

    ExtBoxType::Segment {
        base_offset,
        iv_size,
        n_frames,
        frames
    }
}